"""Scikit-Learn interface for MOVAL

The implementation follows the guide on `Developing scikit-learn estimators`_.

Note:
    The main class of this module, :py:class:`MOVAL`, is also available under the
    top-level package name as :py:class:`moval.MOVAL` and automatically imported
    with :py:mod:`moval`.

.. _Developing scikit-learn estimators:
    https://scikit-learn.org/stable/developers/develop.html
"""

from moval.integrations.sklearn import moval